# coding: utf-8
# Copyright (c) 2015, Truelite Srl <info@truelite.it>
# Author: Mark Caglienzi <mark@truelite.it>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import unicode_literals
import datetime
from django.conf import settings
from django.contrib.auth.models import Permission, User
from django.core.urlresolvers import reverse
from django.db.models.loading import get_model
from django.test import TestCase
from django.test.client import Client
from django.utils.unittest import skip
import django_pettycash.consts as dpc_consts
import django_pettycash.models as dpc_models
import django_pettycash.utils as dpc_utils


# This is a string in the form 'app_name.model_name'
COMPANY_MODEL = getattr(settings, 'COMPANY_MODEL')

DATEFORMAT = getattr(settings, 'DATEFORMAT')

class EntryTests(TestCase):
    """
    Class to test validations in Entry model
    """
    def setUp(self):
        self.client = Client()
        self.user = User(
            username='user',
            is_superuser=False,
        )
        self.user.set_password('user')
        self.user.save()
        # We need a Source
        self.source = dpc_models.Source.objects.create(name="test_source")
        # We need a Company
        # get_model is deprecated in Django 1.7, but
        # https://docs.djangoproject.com/en/1.7/ref/applications/#django.apps.AppConfig.get_model
        # does not work in Django 1.6
        self.company = get_model(
            COMPANY_MODEL.split('.')[0],
            COMPANY_MODEL.split('.')[1],
        ).objects.create(name="test_company")
        self.today = datetime.date.today()
        self.day = datetime.date(
            day=1,
            month=self.today.month,
            year=self.today.year,
        )

    def test_entry_post_save_hook_updates_monthly_balance_with_view(self):
        mb = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_pk = mb.pk
        # Test that the hook works when using the view
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename="add_entry"))
        self.client.login(username='user', password='user')
        response = self.client.post(reverse('dpc_entry_create'), {
            'company': self.company.pk,
            'source': self.source.pk,
            'date': self.today.strftime(DATEFORMAT),
            'kind': dpc_consts.ENTRY_KIND_INBOUND,
            'reason': 'test reason',
            'amount': 10.00,
        })
        self.assertEqual(response.status_code, 302)
        mb = dpc_models.MonthlyBalance.objects.get(pk=mb_pk)
        self.assertEqual(mb.end, 110.00)

    def test_entry_post_save_hook_updates_monthly_balance_without_view(self):
        mb = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_pk = mb.pk
        # Test that the hook works when NOT using the view
        dpc_models.Entry.objects.create(
            company=self.company,
            source=self.source,
            date=self.today,
            kind=dpc_consts.ENTRY_KIND_OUTBOUND,
            reason='test reason',
            amount=10.00,
        )
        mb = dpc_models.MonthlyBalance.objects.get(pk=mb_pk)
        self.assertEqual(mb.end, 90.00)

    def test_entry_post_save_hook_updates_the_right_monthly_balance_with_view(self):
        # Current month MonthlyBalance
        cmb = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        cmb_pk = cmb.pk
        # Previous month MonthlyBalance
        pmb = dpc_models.MonthlyBalance.objects.create(
            date=dpc_utils.get_previous_month(self.day),
            source=self.source,
            start=10.00,
            end=100.00,
        )
        pmb_pk = pmb.pk
        # Test that the hook works when using the view
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename="add_entry"))
        self.client.login(username='user', password='user')
        response = self.client.post(reverse('dpc_entry_create'), {
            'company': self.company.pk,
            'source': self.source.pk,
            'date': self.today.strftime(DATEFORMAT),
            'kind': dpc_consts.ENTRY_KIND_INBOUND,
            'reason': 'test reason',
            'amount': 10.00,
        })
        self.assertEqual(response.status_code, 302)
        cmb = dpc_models.MonthlyBalance.objects.get(pk=cmb_pk)
        self.assertEqual(cmb.end, 110.00)
        pmb = dpc_models.MonthlyBalance.objects.get(pk=pmb_pk)
        self.assertEqual(pmb.end, 100.00)

    def test_entry_post_save_hook_updates_the_right_monthly_balance_without_view(self):
        # Current month MonthlyBalance
        cmb = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        cmb_pk = cmb.pk
        # Previous month MonthlyBalance
        pmb = dpc_models.MonthlyBalance.objects.create(
            date=dpc_utils.get_previous_month(self.day),
            source=self.source,
            start=10.00,
            end=100.00,
        )
        pmb_pk = pmb.pk
        # Test that the hook works when NOT using the view
        dpc_models.Entry.objects.create(
            company=self.company,
            source=self.source,
            date=self.today,
            kind=dpc_consts.ENTRY_KIND_OUTBOUND,
            reason='test reason',
            amount=10.00,
        )
        cmb = dpc_models.MonthlyBalance.objects.get(pk=cmb_pk)
        self.assertEqual(cmb.end, 90.00)
        pmb = dpc_models.MonthlyBalance.objects.get(pk=pmb_pk)
        self.assertEqual(pmb.end, 100.00)

    # refs #5821
    def test_entry_post_save_hook_updates_all_future_monthly_balances_without_view(self):
        dpc_models.MonthlyBalance.objects.all().delete()
        mb_month4 = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month3 = dpc_models.MonthlyBalance.objects.create(
            date=(mb_month4.date - datetime.timedelta(days=1)).replace(day=1),
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month2 = dpc_models.MonthlyBalance.objects.create(
            date=(mb_month3.date - datetime.timedelta(days=1)).replace(day=1),
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month1 = dpc_models.MonthlyBalance.objects.create(
            date=(mb_month2.date - datetime.timedelta(days=1)).replace(day=1),
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month3.delete()
        self.assertEqual(dpc_models.MonthlyBalance.objects.all().count(), 3)
        entry = dpc_models.Entry.objects.create(
            company=self.company,
            source=self.source,
            date=mb_month1.date.replace(day=7),
            kind=dpc_consts.ENTRY_KIND_INBOUND,
            reason='test reason',
            amount=10.00,
        )
        self.assertEqual(entry.amount, 10)
        self.assertEqual(entry.kind, dpc_consts.ENTRY_KIND_INBOUND)
        # (Re)load objects
        mb_month1 = dpc_models.MonthlyBalance.objects.get(pk=mb_month1.pk)
        mb_month2 = dpc_models.MonthlyBalance.objects.get(pk=mb_month2.pk)
        mb_month4 = dpc_models.MonthlyBalance.objects.get(pk=mb_month4.pk)
        # Check objects
        self.assertEqual(mb_month1.start, 100)
        self.assertEqual(mb_month1.end, 110)
        self.assertEqual(mb_month2.start, 110)
        self.assertEqual(mb_month2.end, 110)
        self.assertEqual(mb_month4.start, 110)
        self.assertEqual(mb_month4.end, 110)

    # refs #5821
    def test_entry_post_save_hook_updates_all_future_monthly_balances_with_view(self):
        dpc_models.MonthlyBalance.objects.all().delete()
        mb_month4 = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month3 = dpc_models.MonthlyBalance.objects.create(
            date=(mb_month4.date - datetime.timedelta(days=1)).replace(day=1),
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month2 = dpc_models.MonthlyBalance.objects.create(
            date=(mb_month3.date - datetime.timedelta(days=1)).replace(day=1),
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month1 = dpc_models.MonthlyBalance.objects.create(
            date=(mb_month2.date - datetime.timedelta(days=1)).replace(day=1),
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month3.delete()
        self.assertEqual(dpc_models.MonthlyBalance.objects.all().count(), 3)
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename="add_entry"))
        self.client.login(username='user', password='user')
        response = self.client.post(reverse('dpc_entry_create'), {
            'company': self.company.pk,
            'source': self.source.pk,
            'date': mb_month1.date.replace(day=7).strftime(DATEFORMAT),
            'kind': dpc_consts.ENTRY_KIND_INBOUND,
            'reason': 'test reason',
            'amount': 10.00,
        })
        self.assertEqual(response.status_code, 302)
        # (Re)load objects
        entry = dpc_models.Entry.objects.get(date=mb_month1.date.replace(day=7))
        self.assertEqual(entry.amount, 10)
        self.assertEqual(entry.kind, dpc_consts.ENTRY_KIND_INBOUND)
        mb_month1 = dpc_models.MonthlyBalance.objects.get(pk=mb_month1.pk)
        mb_month2 = dpc_models.MonthlyBalance.objects.get(pk=mb_month2.pk)
        mb_month4 = dpc_models.MonthlyBalance.objects.get(pk=mb_month4.pk)
        # Check objects
        self.assertEqual(mb_month1.start, 100)
        self.assertEqual(mb_month1.end, 110)
        self.assertEqual(mb_month2.start, 110)
        self.assertEqual(mb_month2.end, 110)
        self.assertEqual(mb_month4.start, 110)
        self.assertEqual(mb_month4.end, 110)

    # refs #5824
    def test_entry_post_save_hook_updates_from_the_right_date_in_case_of_modification_without_view(self):
        dpc_models.MonthlyBalance.objects.all().delete()
        mb_month3 = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month2 = dpc_models.MonthlyBalance.objects.create(
            date=(mb_month3.date - datetime.timedelta(days=1)).replace(day=1),
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month1 = dpc_models.MonthlyBalance.objects.create(
            date=(mb_month2.date - datetime.timedelta(days=1)).replace(day=1),
            source=self.source,
            start=100.00,
            end=100.00,
        )
        self.assertEqual(dpc_models.MonthlyBalance.objects.all().count(), 3)
        entry = dpc_models.Entry.objects.create(
            company=self.company,
            source=self.source,
            date=mb_month1.date.replace(day=7),
            kind=dpc_consts.ENTRY_KIND_INBOUND,
            reason='test reason',
            amount=10.00,
        )
        # Reload objects
        mb_month1 = dpc_models.MonthlyBalance.objects.get(pk=mb_month1.pk)
        mb_month2 = dpc_models.MonthlyBalance.objects.get(pk=mb_month2.pk)
        mb_month3 = dpc_models.MonthlyBalance.objects.get(pk=mb_month3.pk)
        # Check objects
        self.assertEqual(mb_month1.start, 100)
        self.assertEqual(mb_month1.end, 110)
        self.assertEqual(mb_month2.start, 110)
        self.assertEqual(mb_month2.end, 110)
        self.assertEqual(mb_month3.start, 110)
        self.assertEqual(mb_month3.end, 110)
        # The real test starts here
        # Edit the entry
        entry.date = mb_month2.date.replace(day=7)
        entry.save()
        # Reload objects
        mb_month1 = dpc_models.MonthlyBalance.objects.get(pk=mb_month1.pk)
        mb_month2 = dpc_models.MonthlyBalance.objects.get(pk=mb_month2.pk)
        mb_month3 = dpc_models.MonthlyBalance.objects.get(pk=mb_month3.pk)
        # Recheck the balances
        self.assertEqual(mb_month1.start, 100)
        self.assertEqual(mb_month1.end, 100)
        self.assertEqual(mb_month2.start, 100)
        self.assertEqual(mb_month2.end, 110)
        self.assertEqual(mb_month3.start, 110)
        self.assertEqual(mb_month3.end, 110)

    # refs #5824
    def test_entry_post_save_hook_updates_from_the_right_date_in_case_of_modification_with_view(self):
        dpc_models.MonthlyBalance.objects.all().delete()
        mb_month3 = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month2 = dpc_models.MonthlyBalance.objects.create(
            date=(mb_month3.date - datetime.timedelta(days=1)).replace(day=1),
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month1 = dpc_models.MonthlyBalance.objects.create(
            date=(mb_month2.date - datetime.timedelta(days=1)).replace(day=1),
            source=self.source,
            start=100.00,
            end=100.00,
        )
        self.assertEqual(dpc_models.MonthlyBalance.objects.all().count(), 3)
        entry = dpc_models.Entry.objects.create(
            company=self.company,
            source=self.source,
            date=mb_month1.date.replace(day=7),
            kind=dpc_consts.ENTRY_KIND_INBOUND,
            reason='test reason',
            amount=10.00,
        )
        # Reload objects
        mb_month1 = dpc_models.MonthlyBalance.objects.get(pk=mb_month1.pk)
        mb_month2 = dpc_models.MonthlyBalance.objects.get(pk=mb_month2.pk)
        mb_month3 = dpc_models.MonthlyBalance.objects.get(pk=mb_month3.pk)
        # Check objects
        self.assertEqual(mb_month1.start, 100)
        self.assertEqual(mb_month1.end, 110)
        self.assertEqual(mb_month2.start, 110)
        self.assertEqual(mb_month2.end, 110)
        self.assertEqual(mb_month3.start, 110)
        self.assertEqual(mb_month3.end, 110)
        # The real test starts here
        # Edit the entry
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename="change_entry"))
        self.client.login(username='user', password='user')
        response = self.client.post(reverse('dpc_entry_update', kwargs={'pk': entry.pk}), {
            'company': entry.company.pk,
            'source': entry.source.pk,
            'date': mb_month2.date.replace(day=7).strftime(DATEFORMAT),
            'kind': entry.kind,
            'reason': entry.reason,
            'amount': entry.amount,
        })
        self.assertEqual(response.status_code, 302)
        # Reload objects
        mb_month1 = dpc_models.MonthlyBalance.objects.get(pk=mb_month1.pk)
        mb_month2 = dpc_models.MonthlyBalance.objects.get(pk=mb_month2.pk)
        mb_month3 = dpc_models.MonthlyBalance.objects.get(pk=mb_month3.pk)
        # Recheck the balances
        self.assertEqual(mb_month1.start, 100)
        self.assertEqual(mb_month1.end, 100)
        self.assertEqual(mb_month2.start, 100)
        self.assertEqual(mb_month2.end, 110)
        self.assertEqual(mb_month3.start, 110)
        self.assertEqual(mb_month3.end, 110)

    @skip("This test does not work because the amount field it's not supposed to allow negative values")
    def test_force_positive_amount(self):
        dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        e = dpc_models.Entry(
            company=self.company,
            source=self.source,
            reason="Reason",
            amount=100,
            date=self.today,
            kind=dpc_consts.ENTRY_KIND_INBOUND,
        )
        e.save()
        self.assertTrue(e.amount > 0)
        e = dpc_models.Entry(
            company=self.company,
            source=self.source,
            reason="Reason",
            amount=-100,
            date=self.today,
            kind=dpc_consts.ENTRY_KIND_INBOUND,
        )
        e.save()
        self.assertTrue(e.amount < 0)
        dpc_models.Entry.objects.all().delete()
